﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Logging;
using ApplePay_SecurePay.Models;

using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Http.Extensions;
using Microsoft.Extensions.Options;
using Newtonsoft.Json;
using System.Security.Cryptography.X509Certificates;
using System.Net;
using System.IO;
using System.Xml;
using System.Text;
using System.Security.Cryptography;
using System.Net.Http;

namespace ApplePay_SecurePay.Controllers
{
    public class HomeController : Controller
    {
        private readonly IOptions<ConfigSetting> config;
        private readonly ILogger<HomeController> _logger;

        public HomeController(IOptions<ConfigSetting> config, ILogger<HomeController> logger)
        {
            this.config = config;
            this._logger = logger;
        }

        public IActionResult Index()
        {
            ViewBag.countryCode = "SG";
            ViewBag.currencyCode = "SGD";
            ViewBag.paymentDescription = "Special Item";
            ViewBag.amount = "10.00";
            return View();

        }

        public IActionResult Privacy()
        {
            return View();
        }

        [ResponseCache(Duration = 0, Location = ResponseCacheLocation.None, NoStore = true)]
        public IActionResult Error()
        {
            return View(new ErrorViewModel { RequestId = Activity.Current?.Id ?? HttpContext.TraceIdentifier });
        }

        #region ApplePay

        public ActionResult validateMerchant()
        {
            return Ok();
        }
        [HttpPost]
        public async Task<ActionResult> ValidateMerchant(AppleValidateMerchantSessionModel appleValidateMerchant)
        {
            _logger.LogInformation($"ValidateMerchant: validationUrl=> {appleValidateMerchant.ValidationUrl}");

            // TODO add additional mapping validation url from Apple Pay
            if (!Uri.TryCreate(appleValidateMerchant.ValidationUrl, UriKind.Absolute, out var requestUri))
            {
                var result = Content("{'err': 'Some error!'}", "application/json; charset=utf-8");
                return BadRequest(new ValidationResponse() { status = "ERROR", data = "Invalid validation url" });
            }

            try
            {
                _logger.LogInformation($"ValidateMerchant: privateKey=> {config.Value.PrivateKeyLocation}");
                X509Certificate2 privateKey = new X509Certificate2(config.Value.PrivateKeyLocation, config.Value.PrivateKeyPassword, X509KeyStorageFlags.MachineKeySet | X509KeyStorageFlags.Exportable);

                AppleMerchantCertificate merchantIdentityCertificate = new AppleMerchantCertificate(privateKey);

                // Create the JSON payload to POST to the Apple Pay merchant validation URL.
                var payloadObj = new AppleMerchantSessionRequest()
                {
                    DisplayName = config.Value.MerchantName,
                    Initiative = "web",
                    InitiativeContext = Request.Host.Host,
                    MerchantIdentifier = merchantIdentityCertificate.GetMerchantIdentifier()
                };
                string payload = JsonConvert.SerializeObject(payloadObj);

                _logger.LogInformation($"Request: url=> {appleValidateMerchant.ValidationUrl}, data=> {payload}");
                string responsePayload;

                var handler = new HttpClientHandler();
                handler.ClientCertificates.Add(merchantIdentityCertificate.GetCertificate());
                _logger.LogInformation($"client certificate added. - " + merchantIdentityCertificate.GetCertificate().FriendlyName);

                var client = new HttpClient(handler);

                var response = await client.PostAsync(appleValidateMerchant.ValidationUrl, new StringContent(payload, Encoding.UTF8));
                if (response.IsSuccessStatusCode)
                {
                    responsePayload = await response.Content.ReadAsStringAsync();
                    _logger.LogInformation($"response data => {responsePayload}");
                    return Json(responsePayload);
                }
                else
                {
                    responsePayload = await response?.Content?.ReadAsStringAsync();
                    _logger.LogError($"ValidateMerchant: httperror: {response.StatusCode}, response: {responsePayload} ");

                    return BadRequest(new ValidationResponse() { status = "ERROR", data = $"Error: {responsePayload}" });
                }
            }
            catch (Exception ex)
            {
                _logger.LogError($"ValidateMerchant: exception=> {ex.Message}, {ex.ToString()}, IE: {ex.InnerException?.ToString()}");

                return BadRequest(new ValidationResponse() { status = "ERROR", data = ex.Message });
            }
        }


        [HttpPost]
        public async Task<ActionResult> Pay(MobileTokenPaymentModel request)
        {
            _logger.LogInformation("START payApplePay");

            string token = request.MobilePaymentToken;


            string url = config.Value.PgwUrl;

            // make payment
            string paymentResponse = string.Empty, error = string.Empty;

            //Merchant Account Information
            string merchantID = "JT01";             //Get MerchantID from 2c2p PGW Dashboard
            string secretKey = "7jYcp4FxFdf0";      //Get SecretKey from 2c2p PGW Dashboard

            //Product Information
            string uniqueTransactionCode = "Invoice" + DateTime.Now.ToString("yyMMddHHmmss");
            string desc = "Test Payment";

            //***** AMOUNT AND CURRENCY WILL GET VALIDATED BY THE API. 
            //      WHEN MISMATCHED BETWEEN APPLEPAY TOKEN and SUBMITTED AMOUNT, PAYMENT WILL NOT BE PROCESSED.

            string amt = "000000001000";            //12 digit format 
            string currencyCode = "702";            //Ref: http://en.wikipedia.org/wiki/ISO_4217



            //Customer Information
            string customerName = "John Doe";
            string country = "SG";

            //Request Information
            string timeStamp = DateTime.Now.ToString("ddMMyyHHmmss");
            string apiVersion = "9.9";

            //Construct payment request message
            string payloadXml = "<PaymentRequest>" +
                    "<timeStamp>" + timeStamp + "</timeStamp>" +
                    "<merchantID>" + merchantID + "</merchantID>" +
                    "<paymentChannel>APPLEPAY</paymentChannel>" +
                    "<uniqueTransactionCode>" + uniqueTransactionCode + "</uniqueTransactionCode>" +
                    "<desc>" + System.Web.HttpUtility.HtmlEncode(desc) + "</desc>" +
                    "<amt>" + amt + "</amt>" +
                    "<currencyCode>" + currencyCode + "</currencyCode>" +
                    "<panCountry>" + country + "</panCountry>" +
                    "<cardholderName>" + customerName + "</cardholderName>" +
                    "<mobilePaymentData>" + request.MobilePaymentToken + "</mobilePaymentData>" +
                    "</PaymentRequest>";

            
            _logger.LogInformation($"PGW Request (RAW): {payloadXml}");
            string payload = base64Encode(payloadXml);                                //Convert payload to base64  
            string hash = getHMAC2(payload, secretKey);                         //Calculate Hash Value


            string xml = "<PaymentRequest>" +
                "<version>" + apiVersion + "</version>" +                                          //request version number (9.9)
                "<payload>" + payload + "</payload>" +                                             //request payload
                "<signature>" + hash + "</signature>" +                                       //signature
            "</PaymentRequest>";


            string data = base64Encode(xml);                                //Convert payload to base64                    

            _logger.LogInformation($"PGW Request (Actual): {data}");


            var response = await PostPayment(data, url);
            if (response.Item1)
            {
                if (!string.IsNullOrEmpty(response.Item2))
                {
                    _logger.LogInformation($"PGW Response (Actual): {response.Item2}");

                    string clearResult = base64Decode(response.Item2);
                    XmlDocument xDocResponse = new XmlDocument();
                    xDocResponse.LoadXml(clearResult);
                    string respPayload = (xDocResponse.SelectNodes("//payload").Count > 0 ? xDocResponse.SelectSingleNode("//payload").InnerText : "");
                    string respSign = (xDocResponse.SelectNodes("//signature").Count > 0 ? xDocResponse.SelectSingleNode("//signature").InnerText : "");

                    string respHash = getHMAC2(respPayload, secretKey);                         //Calculate Hash Value

                    if (respHash.Equals(respSign, StringComparison.OrdinalIgnoreCase))
                    {
                        respPayload = base64Decode(respPayload);
                        _logger.LogInformation($"PGW Response (RAW): {respPayload}");

                        XmlDocument xDocPayload = new XmlDocument();
                        xDocPayload.LoadXml(respPayload);

                        string respCode = (xDocPayload.SelectNodes("//respCode").Count > 0 ? xDocPayload.SelectSingleNode("//respCode").InnerText : "");

                        string paymentStatus = ((respCode == "00") ? "SUCCESS" : "FAILED");
                        return Ok(new { status = paymentStatus });
                    }
                    else
                    {
                        _logger.LogInformation($"Invalid Signature");

                        return BadRequest(new { status = "ERROR", data = "INVALID SIGNATURE" });
                    }
                }
                else
                {
                    return BadRequest(new { status = "ERROR", data = "EMPTY RESULT" });
                }
            }
            else
            {
                return BadRequest(new { status = "ERROR", data = "UNABLE TO POST" });
            };


        }

        #endregion

        private async Task<Tuple<bool,string>> PostPayment(string request, string url)
        {
            string response = "";
            //err = "";
            try
            {
                System.Net.ServicePointManager.SecurityProtocol = SecurityProtocolType.Tls12;

                var client = new HttpClient();

                var payload = new Dictionary<string, string>();
                payload.Add("paymentRequest", request);

                var httpResponse = await client.PostAsync(url, new FormUrlEncodedContent(payload));
                if (httpResponse.IsSuccessStatusCode)
                {
                    response = await httpResponse.Content.ReadAsStringAsync();
                    return new Tuple<bool, string>(true, response);
                }
                else
                {
                    response = await httpResponse?.Content?.ReadAsStringAsync();
                    return new Tuple<bool, string>(false, response);
                }
            }
            catch (Exception ex)
            {
                return new Tuple<bool, string>(false, ex.Message.ToString());
            };
        }


        private string base64Encode(string plainText)
        {
            var plainTextBytes = System.Text.Encoding.UTF8.GetBytes(plainText);
            return Convert.ToBase64String(plainTextBytes);
        }


        private string base64Decode(string encodedText)
        {
            var base64EncodedBytes = System.Convert.FromBase64String(encodedText);
            return Encoding.UTF8.GetString(base64EncodedBytes);
        }

        private string getHMAC2(string signatureString, string secretKey)
        {
            System.Text.UTF8Encoding encoding = new System.Text.UTF8Encoding();
            byte[] keyByte = encoding.GetBytes(secretKey);
            HMACSHA256 hmac = new HMACSHA256(keyByte);
            byte[] messageBytes = encoding.GetBytes(signatureString);
            byte[] hashmessage = hmac.ComputeHash(messageBytes);
            return ByteArrayToHexString(hashmessage);
        }

        private string ByteArrayToHexString(byte[] Bytes)
        {
            StringBuilder Result = new StringBuilder();
            string HexAlphabet = "0123456789ABCDEF";
            foreach (byte B in Bytes)
            {
                Result.Append(HexAlphabet[(int)(B >> 4)]);
                Result.Append(HexAlphabet[(int)(B & 0xF)]);
            }
            return Result.ToString().ToUpper();
        }

    }
}
