﻿using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading.Tasks;

namespace ApplePay_SecurePay.Models
{
    [Serializable]
    public class AppleValidateMerchantSessionModel
    {
        [DataType(DataType.Url)]
        [Required]
        [JsonProperty("validationUrl")]
        public string ValidationUrl { get; set; }
    }
    [Serializable]
    public class AppleMerchantSessionRequest
    {
        [JsonProperty("merchantIdentifier")]
        public string MerchantIdentifier { get; set; }

        [JsonProperty("displayName")]
        public string DisplayName { get; set; }

        [JsonProperty("initiative")]
        public string Initiative { get; set; }

        [JsonProperty("initiativeContext")]
        public string InitiativeContext { get; set; }
    }
    [Serializable]
    public class MobileTokenPaymentModel
    {
        [Required]
        [JsonProperty("mobilePaymentToken")]
        public string MobilePaymentToken { get; set; }
    }

    [Serializable]
    public class AppleMerchantCertificate
    {
        private X509Certificate2 merchantCertificate;

        public AppleMerchantCertificate(string base64PrivateKey, string privateKeyPwd)
        {
            merchantCertificate = LoadCertificate(base64PrivateKey, privateKeyPwd);
        }
        public AppleMerchantCertificate(X509Certificate2 privateKey)
        {
            merchantCertificate = privateKey;
        }

        private AppleMerchantCertificate()
        {
        }

        public X509Certificate2 GetCertificate()
        {
            return merchantCertificate;
        }

        public string GetMerchantIdentifier()
        {
            try
            {
                return GetMerchantIdentifier(merchantCertificate);
            }
            catch (InvalidOperationException)
            {
                return string.Empty;
            }
        }

        private string GetMerchantIdentifier(X509Certificate2 certificate)
        {
            // This OID returns the ASN.1 encoded merchant identifier
            var extension = certificate.Extensions["1.2.840.113635.100.6.32"];

            if (extension == null)
                return string.Empty;

            // Convert the raw ASN.1 data to a string containing the ID
            return Encoding.ASCII.GetString(extension.RawData).Substring(2);
        }

        private X509Certificate2 LoadCertificate(string MerchantCertificateBase64, string MerchantCertificatePassword)
        {
            try
            {
                return new X509Certificate2(Convert.FromBase64String(MerchantCertificateBase64), MerchantCertificatePassword);
            }
            catch (Exception ex)
            {
                throw new InvalidOperationException($"Failed to load Apple Pay merchant certificate.", ex);
            }
        }

    }
}
